Package: expm
Type: Package
Title: Matrix Exponential, Log, 'etc'
Version: 1.0-0
Date: 2024-08-19
Authors@R: c(person("Martin", "Maechler", role=c("aut","cre"),
                    email="maechler@stat.math.ethz.ch", comment = c(ORCID = "0000-0002-8685-9910"))
  , person("Christophe","Dutang", role = "aut", comment = c(ORCID = "0000-0001-6732-1501"))
  , person("Vincent",   "Goulet", role = "aut", comment = c(ORCID = "0000-0002-9315-5719"))
  , person("Douglas",   "Bates",      role = "ctb", comment = "cosmetic clean up, in svn r42")
  , person("David",     "Firth",      role = "ctb", comment = "expm(method= \"PadeO\" and \"TaylorO\")")
  , person("Marina",    "Shapira",    role = "ctb", comment = "expm(method= \"PadeO\" and \"TaylorO\")")
  , person("Michael",   "Stadelmann", role = "ctb", comment = "\"Higham08*\" methods, see ?expm.Higham08...")
  	     )
Contact: expm-developers@lists.R-forge.R-project.org
Description: Computation of the matrix exponential, logarithm, sqrt,
  and related quantities, using traditional and modern methods.
Depends: Matrix
Imports: methods
Suggests: RColorBrewer, sfsmisc, Rmpfr
BuildResaveData: no
License: GPL (>= 2)
URL: https://R-Forge.R-project.org/projects/expm/
BugReports: https://R-forge.R-project.org/tracker/?atid=472&group_id=107
Encoding: UTF-8
NeedsCompilation: yes
Packaged: 2024-08-19 09:24:41 UTC; maechler
Author: Martin Maechler [aut, cre] (<https://orcid.org/0000-0002-8685-9910>),
  Christophe Dutang [aut] (<https://orcid.org/0000-0001-6732-1501>),
  Vincent Goulet [aut] (<https://orcid.org/0000-0002-9315-5719>),
  Douglas Bates [ctb] (cosmetic clean up, in svn r42),
  David Firth [ctb] (expm(method= "PadeO" and "TaylorO")),
  Marina Shapira [ctb] (expm(method= "PadeO" and "TaylorO")),
  Michael Stadelmann [ctb] ("Higham08*" methods, see ?expm.Higham08...)
Maintainer: Martin Maechler <maechler@stat.math.ethz.ch>
Repository: CRAN
Date/Publication: 2024-08-19 15:20:07 UTC
Built: R 4.4.0; amd64-portbld-midnightbsd4; 2026-07-18 06:25:31 UTC; unix
